"""
Functions for parsing embedded metadata from Mars 2020 observational data filenames

This code is derived exclusively from information in
"D-99960 M2020 Camera EDR/RDR Data Products SIS Version 2.0 (Draft)"
(https://pds-imaging.jpl.nasa.gov/reviews/mars2020/mars2020_mission/document_camera/Mars2020_Camera_SIS.pdf)
"""
import itertools
import string
from typing import Literal

import pandas as pd

PType = Literal["IMAGE", "MOSAIC"]
"""Defined M20 imaging product subtypes currently supported by this module."""


def suite(inst):
    raise NotImplementedError


def instrument(fn: str, ptype: PType = "IMAGE") -> str:
    """Return the full instrument name implied by a filename."""
    if ptype == "IMAGE":
        return {
            "FL": "Front Hazcam Left (RCE-A)",
            "FR": "Front Hazcam Right (RCE-A)",
            "FA": "Front Hazcam Anaglyph (RCE-A)",
            "FG": "Front Hazcam Colorglyph (RCE-A)",
            "BL": "Front Hazcam Left (RCE-B)",
            "BR": "Front Hazcam Right (RCE-B)",
            "BA": "Front Hazcam Anaglyph (RCE-B)",
            "BG": "Front Hazcam Colorglyph (RCE-B)",
            "CC": "Cache Cam",
            "EA": "EDL Parachute Uplook Cam A (PUC-A)",
            "EB": "EDL Parachute Uplook Cam B (PUC-B)",
            "EC": "EDL Parachute Uplook Cam C (PUC-C)",
            "ED": "EDL Rover Downlook Cam (RDC)",
            "EL": "EDL Lander Vision System (LVS)",
            "ES": "EDL Descent Stage Downlook Cam (DSD)",
            "EU": "EDL Rover Uplook Cam (RUC)",
            "HN": "Helicopter Navigation Cam",
            "HS": "Helicopter Return To Earth Cam",
            "ZL": "Mastcam-Z Left",
            "ZR": "Mastcam-Z Right",
            "ZA": "Mastcam Anaglyph",
            "ZG": "Mastcam Colorglyph",
            "WS": "MEDA Skycam",
            "NL": "Navcam Left",
            "NR": "Navcam Right",
            "NA": "Navcam Anaglyph",
            "NG": "Navcam Colorglyph",
            "PC": "PIXL Micro Context Cam (MCC)",
            "RL": "Rear Hazcam Left",
            "RR": "Rear Hazcam Right",
            "RA": "Rear Hazcam Anaglyph",
            "RG": "Rear Hazcam Colorglyph",
            "SC": "SHERLOC Context Imager (ACI)",
            "SE": "SHERLOC Engineering (Imagers)",
            "SI": "SHERLOC Imaging (Watson)",
            "SA": "SHERLOC Imaging (Watson) Anaglyph",
            "SG": "SHERLOC Imaging (Watson) Colorglyph",
            "SL": "Watson treated as left-eye",
            "SR": "Watson treated as right-eye",
            "LR": "Supercam RMI",
            "WE": "MEDA Environment",
            "OX": "MOXIE",
            "PE": "PIXL Engineering",
            "PS": "PIXL Spectrometer",
            "LS": "SuperCam Non-Imaging Data",
            "SS": "SHERLOC Spectrometer",
            "XM": "RIMFAX Mobile",
            "XS": "RIMFAX Stationary",
        }[fn[0:2]]
    elif ptype == "MOSAIC":
        return {
            "N": "Navcam",
            "Z": "Mastcam-Z",
            "L": "Supercam RMI",
            "P": "PIXL Micro Context Cam (MCC)",
            "I": "SHERLOC Imaging (WATSON)",
            "C": "SHERLOC ACI",
            "F": "Front Hazcam (RCE-A)",
            "B": "Front Hazcam (RCE-B)",
            "R": "Rear Hazcam",
            "E": "EDL Camera",
            "V": "Mars Helicopter Navigation Cam",
            "X": "Use when there are more than two instruments",
            "_": "N/A",  # Use when there is only one instrument
        }[fn[0:1] if ptype.endswith("1") else fn[1:2]]
    raise NotImplementedError(f"Unsupported filename type: {ptype}")


def color_filter(fn, ptype: PType = "IMAGE") -> str:
    """Return the color filter implied by a filename."""
    # Only “E”, “F”, or “M” or Filter/LED can appear in EDRs.
    if ptype == "IMAGE":
        if "PIXL" in instrument(fn):
            return {
                "R": "LED Red",
                "G": "LED Green",
                "B": "LED Blue",
                "W": "LED Multiple",
                "U": "LED UV",
                "D": "LED SLI-A (Dense)",
                "S": "LED SLI-B (Sparse)",
                "_": "LED Off",
            }[fn[2:3]]
        elif ("Mastcam" in instrument(fn)) and (fn[2:3] in "01234567"):
            return {
                "L0": "L0 (530nm)",
                "L1": "L1 (800nm)",
                "L2": "L2 (754nm)",
                "L3": "L3 (677nm)",
                "L4": "L4 (605nm)",
                "L5": "L5 (528nm)",
                "L6": "L6 (442nm)",
                "L7": "L7 (Solar)",
                "R0": "R0 (530nm)",
                "R1": "R1 (800nm)",
                "R2": "R2 (866nm)",
                "R3": "R3 (910nm)",
                "R4": "R4 (939nm)",
                "R5": "R5 (978nm)",
                "R6": "R6 (1022nm)",
                "R7": "R7 (Solar)",
            }[f"{fn[1:2]}{fn[2:3]}"]
        elif "SHERLOC" in instrument(fn):
            return {
                "0": "cover closed, LEDs off",
                "1": "cover open, LEDs on",
                "2": "cover closed, LEDs off",
                "3": "cover open, LEDs on",
            }[fn[2:3]]
        else:
            return {
                "E": "Raw Bayer pattern",
                "M": "Grayscale",  # (Monochrome/Panchromatic)",
                "A": "Upper Green Bayer",  # cells",
                "D": "Lower Green Bayer",  # cells",
                "O": "Other",
                "_": "N/A",
                "F": "3-Band RGB",
                "R": "Band 1 RGB",
                "G": "Band 2 RGB",
                "B": "Band 3 RGB",
                "T": "3-Band XYZ",
                "X": "Band 1 XYZ",
                "Y": "Band 2 XYZ",
                "Z": "Band 3 XYZ",
                "C": "3-Band xyY",
                "J": "Band 1 xyY",
                "K": "Band 2 xyY",
                "L": "Band 3 xyY",
                "P": "3-Band HSI",
                "H": "Band 1 HSI",
                "S": "Band 2 HSI",
                "I": "Band 3 HSI",
            }[fn[2:3]]
    elif ptype == "MOSAIC":
        return {
            "0": "N/A",
            "_": "N/A",
            "M": "Grayscale",  # (Monochrome/Panchromatic)",
            "U": "UV (PIXL)",
            "F": "3-Band RGB",
            "R": "Band 1 RGB",
            "G": "Band 2 RGB",
            "B": "Band 3 RGB",
            "T": "3-Band XYZ",
            "X": "Band 1 XYZ",
            "Y": "Band 2 XYZ",
            "Z": "Band 3 XYZ",
            "C": "3-Band xyY",
            "J": "Band 1 xyY",
            "K": "Band 2 xyY",
            "L": "Band 3 xyY",
            "P": "3-Band HSI",
            "H": "Band 1 HSI",
            "S": "Band 2 HSI",
            "I": "Band 3 HSI",
        }[fn[3:6]]
    raise ValueError(f"color filters not supported for {ptype}")


def special_flag(fn: str, ptype: PType = "IMAGE") -> str:
    """Return the 'special' flag, if any, implied by a filename"""
    if ptype == "IMAGE":
        return "N/A" if fn[3:4] == "_" else [fn[3:4]]
    elif ptype == "MOSAIC":
        return "Nominal" if fn[6:7] == "_" else fn[6:7]
    raise ValueError(f"Special flags not supported for {ptype}")


def sol(fn, ptype: PType = "IMAGE"):
    """Extract the sol (if any) from a filename."""
    if ptype == "IMAGE":
        if fn[4:8][0] in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":  # Flight or ground test
            return f"{fn[4:8]} - TRANSLATION TO BE DONE"
        else:
            return int(fn[4:8])  # Either flight or test Sol number
    elif ptype == "MOSAIC":
        return "Nominal" if fn[6:7] == "_" else fn[6:7]
    raise NotImplementedError(f"Sol decoding not implemented for {ptype}")


def venue(fn: str, ptype: PType = "IMAGE"):
    """Extract the venue from a filename."""
    if ptype not in ("IMAGE", "MOSAIC"):
        raise NotImplementedError(f"Venue not supported for {ptype}")
    return {
        "_": "Flight",  # (surface or cruise)
        "A": "AVSTB",
        "F": "FSWTB",
        "M": "MSTB",
        "R": "ROASTT",
        "S": "Scarecrow",
        "V": "VSTB",
        "T": "?ORT?",  # Strange option in PDS review data
    }[{"IMAGE": fn[8:9], "MOSAIC": fn[15:16]}[ptype]]


def secondary_timestamp(fn: str) -> int:
    """Extract SCLK (spacecraft clock epoch time) value from a filename"""
    return int(fn[9:19])
    # Note that there is a special calculation to dial this in precisely
    #  for sequences of ZCAM and SHERLOC images which will
    #  all have the same SCLK as the first in the sequence


def tertiary_timestamp(fn: str) -> int:
    """Extract SCLK milliseconds from a filename"""
    return int(fn[20:23])


def product_type(fn: str, ptype: PType = "IMAGE") -> str:
    """Extract instrument-specific product type code from a filename."""
    if ptype not in ("IMAGE", "MOSAIC"):
        raise NotImplementedError(f"product_type not supported for {ptype}")
    val = {"IMAGE": fn[23:26], "MOSAIC": fn[12:15]}[ptype]
    return val


def parse_product_type(fn, ptype="IMAGE"):
    # There is a HUGE lookup table of possible values
    #   which might be implemented here.
    # See: M2020 camera image EDR/RDR product binary formats
    # I have only implemented the ones coded "GREEN" for now
    # (being the standard products to be produced).
    try:
        return {
            "ECM": "Original image product (possibly companded)",
            "EJP": "Original JPEG as recieved from the rover",
            "ECV": "Single frame of a video (possibly companded)",
            "ECZ": "Z-stack frame (possibly companded)",
            "EDM": "Depth map",
            "ECR": "Recovered product (possibly companded)",
            "ERP": "Reference pixel",
            "ERS": "Row summed",
            "ECS": "Column summed",
            "EHG": "Histogram",
            "EAU": "Audio",
            "EDR": "Image (decompanded)",
            "EVD": "Single frame of a video (decompanded)",
            "EZS": "Z-stack frame (decompanded)",
            "ERD": "Recovered product (decompanded)",
            "EBY": "Debayered (demosaicked)",
            "FDR": "Base image from which RDRs are derived",
            "TDR": "Tile data record",
            "FAU": "Audio reconstructed",
            "IOF": "Rad-corrected IOF radiance factor (float)",
            "IOI": "Rad-corrected IOF radiance factor (integer)",
            "RAF": "Rad-corrected (absolute units, float)",
            "RAD": "Rad-corrected (absolute units, 15-bit int static scale)",
            "RAS": "Rad-corrected (absolute units, 12-big int static scale)",
            "RAY": "Rad-corrected (absolute units, 15-bit int dynamic scale)",
            "RDM": "Rad-corrected (absolute units, 15-bit int static scale, masked)",
            "RIE": "Rad corrected for instrument effects only (int DN)",
            "RIF": "Rad-corrected for instrument effects only (float)",
            "RSM": "Rad-corrected (absolute units, 12-big int static scale, masked)",
            "RAG": "Gamma-corrected radiance (no longer calibrated units, byte)",
            "RZF": "Zenith-scaled radiance (float)",
            "RZD": "Zenith-scaled radiance (15-bit int static scale)",
            "RZS": "Zenith-scaled raidance (12-bit int static scale)",
            "RZY": "Zenith-scaled radiance (15-bit int dynamic scale)",
            ### TODO: ALL OF THE "COLOR" PRODUCTS ARE SKIPPED HERE
            "DDD": "Stereo delta disparity (2-band,true disparity offset",
            "DDL": "Stereo delta disparity of lines (single-band)",
            "DDS": "Stereo delta disparity of samples (single-band)",
            "DSE": "Stereo disparity error metric",
            "DSL": "Stereo disparity of lines (single-band)",
            "DSS": "Stereo disparity of sample (single-band)",
            "DSP": "Stereo disparity (final)",
            "DSR": "Stereo disparity (raw)",
            "MDS": "Stereo disparity mask",
            "DFF": "Stereo first-stage disparity (final)",
            "DFL": "Stereo first-stage disparity (line)",
            "DFS": "Stereo first-stage disparity (sample)",
            "XYZ": "XYZ in Site frame",
            "XYM": "XYZ in Site frame (masked)",
            "XYR": "XYZ in Rover Nav frame",
            "ZRM": "XYZ in Rover Nav frame (masked)",
            "XYE": "XYZ error metric",
            "MXY": "XYZ mask",
            "XYF": "XYZ (filled)",
            "XYO": "XYZ w/ overlay",
            "XRZ": "XYZ from disparity raw expressed in Site frame",
            "XRR": "XYZ from disparity raw expressed in Rover Nav frame",
            "XXX": "XYZ X-band",
            "YYY": "XYZ Y-band",
            "ZZZ": "XYZ Z-band",
            "DEM": "Digital elevation model",
            "XOZ": "Orbital XYZ",
            "XXF": "XYZ X-band (filled)",
            "YYF": "XYZ Y-band (filled)",
            "ZZF": "XYZ Z-band (filled)",
            "ZIH": "Instrument Z-value for helicoptor",
            "RNE": "Range error metric",
            "RNF": "Range (filled)",
            "RNG": "Range from camera",
            "RNM": "Range from camera (masked)",
            "RNR": "Range from rover origin",
            "RNO": "Range w/ overlay",
            "UVF": "Surface normal (filled)",
            "UVO": "Surface normal w/ overlay",
            "UVP": "Surface normal projected onto plane",
            "UVS": "Surfaace normal for slope computations",
            "UVT": "Surface normal angle (theta) between normal and plane",
            "UVW": "Surface normal",
            "UUU": "Surface normal U-band",
            "VVV": "Surface normal V-band",
            "WWW": "Surface normal W-band",
            "UUF": "Surface normal U-band filled",
            "VVF": "Surface normal V-band filled",
            "WWF": "Surface normal W-band filled",
            "UIH": "Instrument normal for Helicoptor",
            "TFH": "Tilt for Helicoptor",
            "SEN": "Solar energy",
            "SHD": "Slope heading",
            "SHO": "Slope heading w/ overlay",
            "SLO": "Slope w/ overlay",
            "SLP": "Slope",
            "SMG": "Slope magnitude",
            "SMO": "Slope magnitude w/ overlay",
            "SNO": "Slope northerly tilt w/ overlay",
            "SNT": "Slope northerly tilt",
            "SRD": "Slope radial direction",
            "ARK": "Arm reachability (masked)",
            "ARM": "Arm reachability",
            "ARO": "Arm reachability w/ overlay",
            "MAR": "Arm reachability mask",
            "RUF": "Surface roughness (general, not instrument-specific)",
            "RUS": "Surface roughness (for Drill Stabilizer and Coring)",
            "RUC": "Surface curvature for Drill",
            "RUH": "Surface roughness for Helicoptor",
            "GUN": "Goodness (overall for Natural Surface Tools)",
            "GUA": "Goodness (overall for Abrading Tools)",
            "GUC": "Goodness (overall for Coring Tools)",
            "GRN": "Goodness (for reachability of Natural Surface Tools)",
            "GRA": "Goodness (for reachability of Abrading Tools)",
            "GRC": "Goodness (for reachability of Coring Tools)",
            "GSR": "Goodness (combined for surface roughness)",
            "GUH": "Goodness (Helicoptor)",
            "SLH": "Slope goodness for helicoptor",
            "MSK": "Image mask",
            "TEN": "Terrain class confidence",
            "TER": "Terrain class",
            "IEF": "Incidence, emission, and phase angles (filled)",
            "IEP": "Incidence, emission, and phase angles",
            "IDM": "Index depth map",
            "IDX": "Mosaic input index",
            "ICM": "Mosaic input coregistration map",
        }[product_type(fn, ptype=ptype)]
    except:
        return product_type(fn, ptype=ptype)


def geometry(fn):  # fn[26:27])
    return {
        "_": "N",  # "Non-linearized (raw geometry)",
        "L": "Y",  # "Product has been linearized with nominal stereo partner",
        "A": "Y",  # "Product has been linearized with an actual stereo partner",
    }[fn[26:27]]


def thumbnail(fn):  # fn[27:28])
    return {
        "T": "Y",  # "Product is a thumbnail",
        "N": "N",  # "Nominal Product is a non-thumbnail (full-frame, sub-frame, downsample)",
    }[fn[27:28]]


def site_counter():
    # generates a lookup table per p185 of the SIS
    # 0 - 32767 inclusively, so total 32768
    # last valid is 7DV
    digits, ascii_uppercase = string.digits, string.ascii_uppercase
    collection = (
        itertools.product(digits, repeat=3),
        itertools.product(ascii_uppercase, digits, digits),
        itertools.product(ascii_uppercase, ascii_uppercase, digits),
        itertools.product(ascii_uppercase, repeat=3),
        itertools.product(digits[0:7], ascii_uppercase, ascii_uppercase),
    )
    table = {
        k: v
        for v, k in enumerate(
            map("".join, itertools.islice(itertools.chain(*collection), 32768))
        )
    }
    return table


# greedily evaluate at import time for efficiency
SITE_LOOKUP_TABLE = site_counter()


def site(fn, ptype="IMAGE"):  # fn[28:31])
    val = {"IMAGE": fn[28:31], "MOSAIC": fn[16:19]}[ptype]
    if val == "___":
        return "OUT OF RANGE"
    return SITE_LOOKUP_TABLE[val]


def drive_counter():
    # generates a lookup table per p185-186 of the SIS,
    # 0-65535 inclusively, so total 65536
    # last valid is LJ35
    digits, ascii_uppercase = string.digits, string.ascii_uppercase
    collection = (
        itertools.product(digits, repeat=4),
        itertools.product(ascii_uppercase, digits, digits, digits),
        itertools.product(
            ascii_uppercase[0:12], ascii_uppercase, digits, digits
        ),
    )
    table = {
        k: v
        for v, k in enumerate(
            map("".join, itertools.islice(itertools.chain(*collection), 65536))
        )
    }
    return table


# greedily evaluate at import time for efficiency
DRIVE_LOOKUP_TABLE = drive_counter()


def drive(fn, ptype="IMAGE"):  # fn[31:35])
    val = {"IMAGE": fn[31:35], "MOSAIC": fn[19:23]}[ptype]
    if val == "____":
        return "OUT OF RANGE"
    return DRIVE_LOOKUP_TABLE[val]


def sequence(fn):  # fn[35:44])
    return fn[35:44]


def cam_specific(fn):  # fn[44:48])
    return fn[44:48]  # To do complex reconstruction


def downsample(fn):  # fn[48:49])
    n = int(2 ** int(fn[48:49]))
    return f"{n}x{n}"


def compression(fn):  # fn[49:51])
    if fn[49:51].startswith("I"):
        if int(fn[49:51][1]) < 9:
            return f"{int(fn[49:51][1])} bpp ICER (lossy)"
        else:
            return f">8 bpp ICER (lossy)"
    elif fn[49:51].startswith("L"):
        return {
            "LI": "Lossless (ICER)",
            "LL": "Lossless (LOCO)",
            "LM": "Lossless (Malin)",
            "LU": "Uncompressed",
        }[fn[49:51]]


def producer(fn, ptype="IMAGE"):  # fn[51:52])
    val = {"IMAGE": fn[51:52], "MOSAIC": fn[37:38]}[ptype]
    if val == "J":
        return "JPL"
    elif val == "A":
        return "ASU"
    elif val == "P":
        inst = instrument(fn, ptype="MOSAIC1" if "MOSAIC" in ptype else ptype)
        if ("Hazcam" in inst) or ("Navcam" in inst):  # ECAM
            return "JPL"
        elif "Mastcam" in inst:
            return "ASU (Tempe, AZ)"
        elif inst == "Supercam RMI":
            return "IRAP (France)"
        elif "MCC" in inst:
            return "JPL"
        elif "SHERLOC" in inst:
            return "JPL"
        elif inst == "MEDA Skycam":
            return "Ministry of Education and Science (Spain)"
        elif "EDL" in inst:
            return "JPL"
        elif "Helicopter" in inst:
            return "JPL"
        else:
            return val
    elif val == "_":
        return "undefined/other"
    else:
        return "Instrument Co-I"


def version_counter():
    # generates a lookup table per p189 of the SIS
    # version counter indexing starts at 1
    # first valid 01, last valid ZZ
    digits, ascii_uppercase = string.digits, string.ascii_uppercase
    collection = (
        itertools.islice(
            itertools.product(digits, digits), 1, None
        ),  # 00 is not valid
        itertools.product(ascii_uppercase, digits + ascii_uppercase),
    )
    table = {
        k: v
        for v, k in enumerate(
            map("".join, itertools.chain(*collection)), start=1
        )
    }
    return table


def version(fn, ptype="IMAGE"):  # fn[52:54])
    val = {"IMAGE": fn[52:54], "MOSAIC": fn[38:40]}[ptype]
    if val == "__":
        return "OUT OF RANGE"
    return version_counter()[val]


def ext(fn, ptype="IMAGE"):  # fn[54:58])
    try:
        return {
            "VIC": "VICAR",
            "IMG": "VIC w/ ODL label",
            "TIF": "TIFF",
            "JPG": "JPEG",
            "PNG": "PNG",
            "TXT": "ASCII",
            "iv": "Inventor-format",
            "ht": "VICAR (Height-map)",
            "rgb": "SGI RGB (Skin)",
            "obj": "Wavefront OBJ (Mesh)",
            "mtl": "OBJ (Material)",
            "png": "Browse image -or- Mesh",  # ???
            "mlp": "MeshLab project",
            "xml": "PDS label in xml format",
            "xmlf": "filter file w/ rover mask polygons in xml format",
            "WAV": "WAV-format audio",
            "flac": "Free Lossless Audio",
            # "MXML" : "masked XYZ (MXY) xml", # Deprecated?
            "png.xml": "Browse image xml format",  # ????
        }[
            {
                "IMAGE": fn.split(".", 1)[-1],  # fn[55:],
                "MOSAIC:": fn.split(".", 1)[-1],  # fn[42:]
            }[ptype]
        ]
    except KeyError:
        return f"Unknown: {fn.split('.',1)[-1]}"


def get_ptype(fn):  # get the product type
    try:
        return {54: "IMAGE", 40: "MOSAIC"}[
            len(fn.split(".")[0])
        ]  # cut off the extension
    except:
        return "UNKNOWN"


def parse(fn):
    ptype = get_ptype(fn)
    if not ptype in ["IMAGE", "MOSAIC", "MESH"]:
        print(f"Unknown filename type.")
        return
    elif ptype == "MESH":
        print("MESH filenames are not yet supported.")
        return
    # All the ugly conditionals are just to order the information in the
    # most useful way for display.
    fntable = {}
    fntable["FILENAME"] = fn
    if ptype == "IMAGE":
        fntable["INSTRUMENT"] = instrument(fn, ptype=ptype)
    elif ptype == "MOSAIC":
        fntable["INSTRUMENT1"] = instrument(fn, ptype=ptype + "1")
        fntable["INSTRUMENT2"] = instrument(fn, ptype=ptype + "2")
    fntable["FILTER"] = color_filter(fn, ptype=ptype)
    fntable["SOL"] = sol(fn, ptype=ptype)
    fntable["PRODUCT_TYPE"] = product_type(fn, ptype=ptype)
    if ptype == "IMAGE":
        fntable["SCLK"] = (
            secondary_timestamp(fn) + tertiary_timestamp(fn) / 1000
        )
        fntable["SEQUENCE"] = sequence(fn)
        fntable["LINEARIZED"] = geometry(fn)
        fntable["THUMBNAIL"] = thumbnail(fn)
        # fntable["CAM_SPECIFIC"] = cam_specific(fn)  ###
        fntable["DOWNSAMPLED"] = downsample(fn)
        fntable["COMPRESSION"] = compression(fn)
    # fntable["SPECIAL"] = special_flag(fn,ptype=ptype)
    fntable["VENUE"] = venue(fn, ptype=ptype)
    fntable["SITE"] = site(fn, ptype=ptype)
    fntable["DRIVE"] = drive(fn, ptype=ptype)
    fntable["PRODUCER"] = producer(fn, ptype=ptype)
    fntable["VERSION"] = version(fn, ptype=ptype)
    fntable["FILETYPE"] = ext(fn, ptype=ptype)
    fntable["PRODUCT_DESCRIPTION"] = parse_product_type(fn, ptype=ptype)

    return pd.Series(fntable)


class Parse:
    def __call__(self, fn):
        # TODO: A decision tree between missions. Right now, everything is
        #  M2020.
        return parse(fn)



"""
This module contains functions for performing transformations between
coordinate systems. It assumes that the coordinate systems of interest are
defined in relation to one another by offsets and/or rotation quaternions. It
also contains utilities for preprocessing VICAR-style coordinate system
definitions from PDS3 labels in order to facilitate easy transformations from,
e.g., the SOME_INSTRUMENT frame to the SITE frame to the SOME_OTHER_INSTRUMENT
frame.
"""
from itertools import product
from functools import reduce
from numbers import Real
from operator import or_
import re
from typing import Optional, Union, Literal, Sequence

from dustgoggles.func import is_it
from dustgoggles.structures import NestingDict, get_from
import numpy as np
from numpy.typing import ArrayLike
import pandas as pd
import pdr

Quaternion = Sequence[Real]
"""
Simple alias to clarify type hints. Denotes a quaternion expressed in the form
used by functions in this module. (Note that all functions in this module that
return a Quaternion return it as an ndarray, but will accept both ndarrays 
and other types of sequences as arguments.)

An argument annotated as Quaternion should always be a size-4 1D sequence of 
real numbers in which the first element represents the quaternion's scalar part
and the remaining elements represent its basis/vector part. 

In other words:

`quat: Quaternion = np.array([a, b, c, d])` represents the quaternion (a, v),
where a is a real number and v is the element [b, c, d] of ℝ3 (the real 
coordinate space of dimension 3). 

Equivalently, `quat` represents a + b * i + c * j + d * k, where (i, j, k) are 
the basis elements of a real vector space of dimension 3.
"""

UnitOfAngle = Literal["degrees", "radians", "deg", "rad"]


# TODO: this may be cruft.
def get_geometry_value(
    entity_name: str, frame_name: str, axis_name: str, data: pdr.Data
) -> Union[float, int, str]:
    """
    Fetch the value named `entity_name` for the axis named `axis_name` of the
    coordinate frame named `frame_name`, as defined in the metadata of `data`.
    """
    return get_from(
        data.metaget(f"{frame_name}_DERIVED_GEOMETRY_PARMS"),
        (f"{entity_name}_{axis_name}", "value"),
    )


# TODO: these functions feel messy and redundant.
def get_coordinate_system_properties(
    frame_name: str, data: pdr.Data
) -> Optional[dict[str, Union[np.ndarray, str, float]]]:
    """
    Fetch and format information about a VICAR-style coordinate system named
    `frame_name` from a PDS3-labeled product.
    If there is no coordinate system named `frame_name`, returns None.
    """
    syskeys = [
        k
        for k in data.metadata.fieldcounts
        if re.match(f"{frame_name}_COORD(INATE)?_SYSTEM(_PARMS)?$", k)
    ]
    if len(syskeys) == 0:
        return None
    system = data.metaget(syskeys[0])
    return {
        "name": system.get("COORDINATE_SYSTEM_NAME"),
        "reference_frame": system.get("REFERENCE_COORD_SYSTEM_NAME"),
        "quaternion": np.array(system.get("ORIGIN_ROTATION_QUATERNION")),
        "offset": np.array(system.get("ORIGIN_OFFSET_VECTOR")),
        "orientation": system.get("POSITIVE_AZIMUTH_DIRECTION"),
    }


def get_coordinate_systems(
    data: pdr.Data,
) -> dict[str, dict[str, Union[np.ndarray, str, float]]]:
    """
    Fetch and format information about all VICAR-style coordinate systems
    defined in a PDS3-labeled product.
    """
    # TODO: messy
    frame_names = [
        k.split("_")[0]
        for k in data.metadata.fieldcounts
        if re.match(r"\w+_COORD(INATE)?_SYSTEM(_PARMS)?$", k)
    ]
    return {
        name: get_coordinate_system_properties(name, data)
        for name in frame_names
    }


def _check_unit_of_angle(unit: str) -> None:
    """
    Integrity check  for angular coordinate representation conversion
    functions. Freaks out if we don't know anything about a specified unit of
    angle.
    """
    if unit not in {"degrees", "radians", "deg", "rad"}:
        raise TypeError(
            f"Unsupported angular unit {unit}. Pass 'degrees'/'deg' or "
            f"'radians'/'rad' (default is degrees)."
        )


def cart2sph(
    x0: ArrayLike, y0: ArrayLike, z0: ArrayLike, unit: UnitOfAngle = "degrees"
) -> Union[pd.DataFrame, tuple[Real, Real, Real]]:
    """
    Classic Cartesian-to-spherical coordinate representation converter.
    Returns latitude and longitude in degrees by default.
    Pass `unit="radians"` to return radians. If any of `x0`, `y0`, or `z0` are
    pandas Series/DataFrames or numpy ndarrays, returns a DataFrame with
    columns "lat", "lon", and "r". Otherwise, returns a tuple like
    (lat, lon, r).

    Notes:
        1. Assumes that latitude runs from -90 to 90 degrees.
        2. Returns strictly positive longitude (i.e., uses a 0-360 degree
           longitude system).
        3. If more than one of x0, y0, or z0 are of Series/DataFrame/ndarray
           types, all such arguments must have the same shape. (You may,
           however, mix scalars in as you please.)
    """
    _check_unit_of_angle(unit)
    radius = np.sqrt(x0 ** 2 + y0 ** 2 + z0 ** 2)
    if x0 != 0:
        longitude = np.arctan2(y0, x0)
    else:
        longitude = np.pi / 2
    longitude = longitude % (np.pi * 2)
    latitude = np.arcsin(z0 / np.sqrt(x0 ** 2 + y0 ** 2 + z0 ** 2))
    if unit in {"degrees", "deg"}:
        latitude = np.degrees(latitude)
        longitude = np.degrees(longitude)
    if reduce(
        or_, map(is_it(pd.DataFrame, np.ndarray, pd.Series), [x0, y0, z0])
    ):
        return pd.DataFrame({"lat": latitude, "lon": longitude, "r": radius})
    return latitude, longitude, radius


def sph2cart(
    lat: ArrayLike,
    lon: ArrayLike,
    radius: ArrayLike = 1,
    unit: UnitOfAngle = "degrees",
) -> Union[pd.DataFrame, tuple[Real, Real, Real]]:
    """
    Classic spherical-to-Cartesian coordinate representation converter. By
    default, assumes `lat` and `lon` are in degrees; pass `unit="radians"` if
    they are in radians. If any of `lat`, `lon`, or `radius` are pandas
    Series/DataFrames or numpy ndarrays, returns a DataFrame with columns "x",
    "y", and "z". Otherwise, returns a tuple like (x, y, z).

    caveats:
    1. this assumes a coordinate convention in which latitude runs from -90
        to 90 degrees.
    """
    # TODO: do there need to be other caveats?
    _check_unit_of_angle(unit)
    if unit in {"degrees", "deg"}:
        lat = np.radians(lat)
        lon = np.radians(lon)
    x0 = radius * np.cos(lat) * np.cos(lon)
    y0 = radius * np.cos(lat) * np.sin(lon)
    z0 = radius * np.sin(lat)

    if reduce(
        or_,
        map(is_it(pd.DataFrame, np.ndarray, pd.Series), [lat, lon, radius]),
    ):
        return pd.DataFrame({"x": x0, "y": y0, "z": z0})
    return x0, y0, z0


# TODO, maybe: this isn't a very good name
def quaternion_multiplication(q1: Quaternion, q2: Quaternion) -> Quaternion:
    """
    Implements the conventional quaternion multiplication operation, i.e.,
    returns q1 ⋅ q2 for q1, q2 ∈ ℍ. See `Quaternion` docstring for format
    convention.
    """
    q1, q2 = map(np.asarray, (q1, q2))
    s1, v1, s2, v2 = q1[0], q1[1:], q2[0], q2[1:]
    scalar = s1 * s2 - np.dot(v1, v2)
    vector = s1 * v2 + s2 * v1 + np.cross(v1, v2)
    return np.array([scalar, *vector])


def invert_quaternion(quat: Quaternion) -> Quaternion:
    """
    Return the quaternion whose vector part is the additive inverse of `quat`'s
    vector part and whose scalar part is `quat`'s scalar part. For example,
    `invert_quaternion([1, 2, 3, 4])` returns `[1, -2, -3, -4]`.
    """
    return np.array([quat[0]] + [-1 * e for e in quat[1:]])


def rotate_unit_vector(
    alt: float, az: float, quat: Quaternion, clockwise: bool = True
) -> np.ndarray:
    """
    Apply a rotation expressed as a unit quaternion to a unit vector expressed
    as spherical coordinates.

    Assumes units of degrees. Also assumes left-handed coordinate systems by
    default; pass `clockwise=False` to specify right-handed system.
    """
    source_cartesian = np.array(sph2cart(alt, az))
    if clockwise is True:
        source_cartesian *= np.array([-1, -1, 1])
    zero_quaternion = np.array([0, *source_cartesian])
    inverse_rotation = invert_quaternion(quat)
    q_times_0_v = quaternion_multiplication(quat, zero_quaternion)
    v_prime = quaternion_multiplication(q_times_0_v, inverse_rotation)
    assert np.isclose(v_prime[0], 0)
    target_cartesian = v_prime[1:]
    if clockwise is True:
        target_cartesian *= np.array([-1, -1, 1])
    # TODO: it's pointless to return the third (radius) element, because
    #  the result will always be a unit vector. Will probably need to modify
    #  downstream to make this change.
    return np.array(cart2sph(*target_cartesian))


def get_coordinates(
    data: pdr.Data,
) -> dict[str, dict[str, dict[Literal["AZIMUTH", "ELEVATION"], float]]]:
    """
    Fetch and organize all VICAR-style azimuth/elevation values mentioned in
    the metadata of a pdr.Data object.

    For example, if a product's label gives INSTRUMENT and SOLAR
    azimuth/elevation values in both SITE and ROVER frames:

    >>> coordinated_data = pdr.read("coordinated.lbl")
    >>> get_coordinates(coordinated_data)

    Expected output (numeric values will of course vary depending on product):
    ```
    {'ROVER': {'INSTRUMENT': {'AZIMUTH': 351.484, 'ELEVATION': -44.5026},
    'SOLAR': {'AZIMUTH': 89.4054, 'ELEVATION': 67.2452}},
    'SITE': {'INSTRUMENT': {'AZIMUTH': 175.589, 'ELEVATION': -35.0211},
    'SOLAR': {'AZIMUTH': 293.467, 'ELEVATION': 63.6488}}}
    ```
    """
    # TODO: maybe expand this
    axes = ("AZIMUTH", "ELEVATION")
    entities = set()
    for axis in axes:
        entities.update(
            {
                k.replace(f"_{axis}", "")
                for k in data.metadata.fieldcounts
                if k.endswith(f"_{axis}")
            }
        )
    coordinates = NestingDict()
    syskeys = filter(
        lambda k: k.endswith("_GEOMETRY_PARMS"), data.metadata.fieldcounts
    )
    for key, axis, entity in product(syskeys, axes, entities):
        block = data.metaget(key)
        if block is None:
            continue
        record = block.get(f"{entity}_{axis}")
        if record is not None:
            coordinates[key.split("_")[0]][entity][axis] = record["value"]
    return coordinates.todict()


def transform_angle(
    source_frame: str, target_frame: str, entity: str, data: pdr.Data
) -> np.ndarray:
    """
    Transform an angle described in VICAR-style notation in a PDS3 label from
    one coordinate system to another. The label must also provide a quaternion
    defining the relative orientations of those coordinate systems.

    NOTE: "Angle" here is slightly misleading; it really refers to a unit
      direction vector expressed as altitude/azimuth coordinates.

    For example, assuming that "coordinated.lbl" defines both SOLAR_AZIMUTH
    and SOLAR_ELEVATION in SITE frame, and also gives a quaternion that
    rotates coordinates in SITE to coordinates in HEAD, this will return solar
    altitude and azimuth in HEAD:

    >>> coordinated_data = pdr.read("coordinated.lbl")
    >>> transform_angle("SITE", "HEAD", "SOLAR", coordinated_data)
    """
    coordinates = get_coordinates(data)
    systems = get_coordinate_systems(data)
    source_info = systems.get(source_frame)
    target_info = systems.get(target_frame)
    if source_info is None:
        if not target_info["reference_frame"].startswith(source_frame):
            raise ValueError("can't directly convert between these frames.")
        quaternion = invert_quaternion(target_info["quaternion"])
    else:
        quaternion = source_info["quaternion"]
    coord = coordinates[source_frame][entity]
    if target_info is not None:
        clockwise = "clockwise" in target_info["orientation"].lower()
    else:
        clockwise = "clockwise" in source_info["orientation"].lower()
    return rotate_unit_vector(
        coord["ELEVATION"], coord["AZIMUTH"], quaternion, clockwise
    )



import warnings

import numpy as np
from hypothesis import given, strategies as st, assume
from hypothesis.extra.numpy import arrays, array_shapes

from marslab.imgops.render import decorrelation_stretch
from marslab.tests.utilz.utilz import positive_finite_floats, finite_floats
from marslab.tests.utilz.div0 import divide_by_zero


@given(
    channels=st.lists(
        arrays(
            dtype=np.float32,
            shape=st.shared(
                array_shapes(min_dims=2, max_dims=2, min_side=2), key="hi"
            ),
            elements={
                "allow_nan": False,
                "allow_infinity": False,
                "min_value": -1e8,
                "max_value": 1e8,
            },
            fill=st.nothing(),
        ),
        min_size=3,
        max_size=3,
    ),
    contrast_stretch=st.one_of(
        st.none(),
        st.floats(min_value=0, max_value=100),
        st.lists(
            st.floats(min_value=0, max_value=100), max_size=2, min_size=2
        ),
    ),
    special_constants=st.one_of(st.none(), st.lists(finite_floats)),
    sigma=st.one_of(st.none(), positive_finite_floats),
)
def test_fuzz_decorrelation_stretch_nice(
    channels, contrast_stretch, special_constants, sigma
):
    with warnings.catch_warnings():
        divide_by_zero()
        # single-valued arrays have undefined covariance
        assume(all([not len(np.unique(channel)) == 1 for channel in channels]))
        # ridiculously large inputs relative to dtype we expect to cause errors
        assume(all([not np.isinf(np.std(channel)) for channel in channels]))
        if special_constants is not None:
            channels = [
                np.ma.masked_where(np.isin(c, special_constants), c)
                for c in channels
            ]
        decorrelation_stretch(
            channels=channels, contrast_stretch=contrast_stretch, sigma=sigma
        )


# @hset(max_examples=500, verbosity=Verbosity.verbose)
# @given(
#     channels=st.lists(
#         arrays(
#             dtype=st.one_of(floating_dtypes(), integer_dtypes()),
#             shape=st.shared(
#                 array_shapes(min_dims=2, max_dims=2, min_side=2), key="hi"
#             ),
#             elements={
#                 "allow_nan": False,
#                 "allow_infinity": False,
#             },
#             fill=st.nothing(),
#         ),
#         min_size=2,
#     ),
#     contrast_stretch=st.one_of(
#         st.none(),
#         st.floats(min_value=0, max_value=100),
#         st.lists(
#             st.floats(min_value=0, max_value=100), max_size=2, min_size=2
#         ),
#     ),
#     special_constants=st.one_of(
#         st.none(),
#         st.lists(finite_floats),
#         st.sets(finite_floats),
#         st.frozensets(st.floats()),
#         st.dictionaries(keys=finite_floats, values=finite_floats),
#         st.dictionaries(keys=finite_floats, values=st.none()).map(dict.keys),
#         st.dictionaries(keys=st.integers(), values=finite_floats).map(
#             dict.values
#         ),
#         st.dictionaries(keys=finite_floats, values=finite_floats).map(
#             ChainMap
#         ),
#     ),
#     sigma=st.one_of(st.none(), positive_finite_floats),
# )
# def test_fuzz_decorrelation_stretch_nasty(
#     channels, contrast_stretch, special_constants, sigma
# ):
#     # try:
#     # single-valued arrays have undefined covariance
#     assume(all([not len(np.unique(channel)) == 1 for channel in channels]))
#     # ridiculously large inputs relative to dtype we expect to cause errors
#     assume(all([not np.isinf(np.std(channel)) for channel in channels]))
#
#     decorrelation_stretch(
#         channels=channels,
#         contrast_stretch=contrast_stretch,
#         special_constants=special_constants,
#         sigma=sigma,
#     )



from pathlib import Path

import pandas as pd


def parse_m3_fn(fn):
    metadata = {}
    if "m3g" in fn.lower():
        metadata["mode"] = "global"
    elif "m3t" in fn.lower():
        metadata["mode"] = "target"
    if "l0" in fn.lower():
        metadata["level"] = "l0"
    elif "l1b" in fn.lower():
        metadata["level"] = "l1b"
    elif "l2" in fn.lower():
        metadata["level"] = "l2"
    return metadata


DATA_PATH = Path(Path(__file__).parent, "data")

TARGET_WAVELENGTHS = pd.read_csv(Path(DATA_PATH, "m3_target_wavelengths.csv"))
GLOBAL_WAVELENGTHS = pd.read_csv(Path(DATA_PATH, "m3_global_wavelengths.csv"))



"""skeleton BandSet implementation for Clementine mosaics"""

from functools import partial
from pathlib import Path

import pandas as pd

from marslab.compat.clementine import (
    WAVELENGTH_TO_FILTER,
    MOSAIC_SPECIAL_CONSTANTS,
)
from marslab.bandset.bandset import BandSet
from marslab.imgops.loaders import pdr_load


def setup_clem_bandset_metadata(mosaic_file_path):
    dataset = Path(mosaic_file_path).name.split("_")[0]
    row_template = {"PATH": mosaic_file_path, "DATASET": dataset}
    rows = []
    for ix, wave_filter in enumerate(WAVELENGTH_TO_FILTER[dataset].items()):
        row = row_template.copy()
        row["BAND"] = wave_filter[1]
        row["WAVELENGTH"] = wave_filter[0]
        row["IX"] = ix
        rows.append(row)
    return pd.DataFrame(rows)


class ClemBandSet(BandSet):
    """
    a BandSet made from a tile from the Clementine NIR or UVVIS mosaics.
    """

    def __init__(self, tile_path, rois=None, threads=None):
        metadata = setup_clem_bandset_metadata(tile_path)
        load_method = partial(pdr_load, object_name="image")
        super().__init__(
            metadata=metadata,
            load_method=load_method,
            rois=rois,
            threads=threads,
        )
